(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_window.ml,v 1.16 2014-02-26 00:12:42 fclement Exp $ *)

(* The graphics window is split into sub-windows. Each sub-window is
 surrounded by a white margin.
 
 Minimal_graphics -> 1x1 sub-window,
 Parameter_graphics -> 4x1 or 1x4 sub-windows,
 Compound_graphics -> 3x1 or 1x3 sub-windows,
 Full_graphics -> 4x3 or 3x4 sub-windows.
 
 Minimal_graphics means that only the current segmented image is displayed.
 Hence, only one sub-window.
 
 Parameter_graphics means that the three Red/Green/Blue channels, and the
 RGB superimposition, of the current segmented image are displayed.
 Hence, four sub-windows. There is an additional separation of the double
 size of the margin between the three channels and the superimposition.
 
 Compound_graphics means that the data image, the current segmented image,
 and the current gradient image are displayed.
 Hence, three sub-windows. There is an additional separation of the double
 size of the margin between data image and current segmented image rows.
 
 Full_graphics means that the 3+1 components for the data image, the current
 segmented image, and the current gradient image are displayed.
 Hence, 12 sub-windows.
 
 Depending on the aspect ratio of the data image, components sub-windows are
 either displyed along lines or columns:
 for Parameter_graphics,
   Portrait -> 4x1 sub-windows
   Landscape -> 1x4 sub-windows,
 for Compound_graphics,
   Portrait -> 3x1 sub-windows
   Landscape -> 1x3 sub-windows,
 for Full_graphics,
   Portrait -> 4x3 sub-windows
   Landscape -> 3x4 sub-windows. *)

module Nat = Unatural;;
module Prational = Upositive_rational;;
module Prat_range = Upositive_rational_range;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
and _4_ = Nat._4_
;;

open Upositive_rational_range_types;;

let dw_qual = Upervasives.qualify "Display_window";;

let dw_invalid_arg = Upervasives.invalid_arg_qual dw_qual;;

open Configuration_types;;
open Display_types;;

(* Compound graphics window. *)

let aspect (w, h) = if w <= h then Portrait else Landscape;;

let rgb_direction sz =
  match aspect sz with
  | Portrait -> Rgb_Horizontal
  | Landscape -> Rgb_Vertical
;;

let same_direction dir rgb_dir =
  match (dir, rgb_dir) with
  | (Horizontal, Rgb_Horizontal) | (Vertical, Rgb_Vertical) -> true
  | (Horizontal, Rgb_Vertical) | (Vertical, Rgb_Horizontal) -> false
;;

(* See description of the contents of the graphics window in the header. *)
let coeffs gk dir rgb_dir tm =
  let b = same_direction dir rgb_dir in
  let one = (_1_, _0_) and three = (_3_, tm) and four = (_4_, tm) in
  match gk with
  | (Param_quantity, Compound_channel) -> one
  | (Param_quantity, All_channels) -> if b then four else one
  | (All_quantities, Compound_channel) -> if b then three else one
  | (All_quantities, All_channels) -> if b then four else three
;;

(* nn = alpha*(l + 2*m) + beta. *)
let compound_length gk dir rgb_dir m l =
  let two_ms = Nat.mul _2_ m in
  let (alpha, beta) = coeffs gk dir rgb_dir two_ms in
  let sub_length = Nat.add l two_ms in
  Nat.add (Nat.mul alpha sub_length) beta
;;

(* Solve compound formula with l <- r*l for r.
 r = (nn - beta - alpha*2*m) / (alpha*l). *)
let split_ratio gk dir rgb_dir m l nn =
  let two_ms = Nat.mul _2_ m in
  let (alpha, beta) = coeffs gk dir rgb_dir two_ms in
  let p = Nat.sub (Nat.sub nn beta) (Nat.mul alpha two_ms)
  and q = Nat.mul alpha l in
  Prational.of_nat p q
;;

let margin = Nat.of_int 10;;

let scaling_factor gk (nnx_min, nny_min) (nnx_max, nny_max) (w, h) =
  let rgb_dir = rgb_direction (w, h) in
  let rgx =
    Prat_range.of_rat
      (split_ratio gk Horizontal rgb_dir margin w nnx_min)
      (split_ratio gk Horizontal rgb_dir margin w nnx_max)
  and rgy =
    Prat_range.of_rat
      (split_ratio gk Vertical rgb_dir margin h nny_min)
      (split_ratio gk Vertical rgb_dir margin h nny_max) in
  let rg = Prat_range.intersection rgx rgy in
  match (rgx, rgy, rg) with
  | (Empty, _, _) | (_, Empty, _) ->
      dw_invalid_arg "scaling_factor" "incompatible window size bounds"
  | (Nonempty rgx, Nonempty rgy, Empty) ->
      Prational.min rgx.right rgy.right
  | (Nonempty _, Nonempty _, Nonempty _) ->
      Prat_range.project rg Prational._1_
;;

(* Graphics_elements. *)

let init_graphics_elements =
  {
   rgb_direction = Rgb_Horizontal;
   scaling_factor = Prational._1_;
   display_image_size = (Nat._0_, Nat._0_);
   window_size = (Nat._0_, Nat._0_);
  }
;;

(* This should be an abstract reference with unique assignment, weakening it to
 mandatory assignment is necessary for testing purposes though... *)
let { Uaref.get = get_graphics_elements;
      set = gen_set_graphics_elements; } =
  Uaref.make_aref_mandatory "graphics elements" init_graphics_elements
;;

let set_graphics_elements gk min_w_sz max_w_sz ((w, h) as sz) =
  let rgb_dir = rgb_direction (w, h) in
  let sf = scaling_factor gk min_w_sz max_w_sz sz in
  let scale = Prational.scale sf in
  let ((nx, ny) as d_sz) = (scale w, scale h) in
  let nnx = compound_length gk Horizontal rgb_dir margin nx
  and nny = compound_length gk Vertical rgb_dir margin ny in
  let ge =
    {
     rgb_direction = rgb_dir;
     scaling_factor = sf;
     display_image_size = d_sz;
     window_size = (nnx, nny);
    } in
  gen_set_graphics_elements ge
;;

(* Display window setup. *)

let open_display (nnx, nny) =
  Graphics.open_graph "";
  Graphics.resize_window (Nat.to_int nnx) (Nat.to_int nny);
;;

let window_setup gk min_w_sz max_w_sz sz =
  set_graphics_elements gk min_w_sz max_w_sz sz;
  let { window_size = w_sz; _; } = get_graphics_elements () in
  open_display w_sz;
;;
